# -*- coding: utf-8 -*-

"""
    chru.web.routes handles the Flask app.

    This module contains all of the ``@app.route`` calls,
    and it sets up the module wide settings, so the rest of
    the app can use them without problems.
"""

import json
import re
import time
from datetime import datetime
import logging

from flask import Flask, render_template, session, abort, request, g
from flask import redirect, url_for as flask_url_for, flash, make_response, jsonify

from recaptcha.client import captcha

# ----- #
import chru
import chru.utility as utility
import chru.web as web
import chru.web.slug
# API versions
import chru.api.web.v1
from chru.utility.funcs import real_ip, url_for

reserved_routes = ["index", "submit"]

app = Flask(__name__)
app.jinja_env.globals.update(sorted=sorted)
app.jinja_env.globals.update(len=len)
app.jinja_env.globals.update(date_strip_day=utility.funcs.date_strip_day)

def set_app(debug=False, settings_file=None, log_file=None, __supported__=[], update_schema=False):
    """ Sets up the finer details of the Flask app,
        as well as the parsing in the settings file.

        :param debug: should the app be set to debug mode?
        :type debug: bool
        :param settings_file: the file handler for the given settings file
        :type settings_file: file handler
        :param log_file: the file we're going to set :mod:`logging` to use
        :type log_file: str
        :param __supported__: the list of supported settings versions
        :type __supported__: list
        :param update_schema: shall we forcefully update the database schema?
        :type update_schema: bool
    """
    settings_dict = json.load(settings_file)
    settings_file.close()
    settings = utility.struct(**settings_dict)

    web.s = settings
    web.s.debug = debug
    web.s.reserved = reserved_routes + web.s.reserved

    if "_version" not in settings_dict \
            or web.s._version not in __supported__:
        v = "unknown"
        if "_version" in settings_dict:
            v = settings_dict["_version"]
        import sys
        print "Given settings file (v: {0}) is not supported!".format(v)
        print "Supported:", ", ".join(__supported__)
        print "Please make a new one (--make-config > settings.json) and start again."
        return False

    settings_problems = []
    settings_warnings = []
    for k, v in web.skeleton.items():
        if k not in settings_dict:
            settings_problems.append("Missing required field: '{0}'".format(k))
    for k, v in settings_dict.items():
        if k not in web.skeleton:
            settings_warnings.append("Unneeded field: '{0}'".format(k))

    if settings_problems or settings_warnings:
        print "There are some problems with your settings file:"
    if settings_warnings:
        print "\n".join(settings_warnings)
    if settings_problems:
        print "\n".join(settings_problems)
        print "Please fix these and try again"
        return False

    utility.funcs.verify_sql(force=update_schema)

    app.debug = debug
    app.secret_key = str(web.s.flask_secret_key)

    logging.basicConfig(
        format="%(asctime)s [%(levelname)s] %(message)s",
        datefmt="%m/%d/%Y %I:%M:%S%p",
        level=logging.DEBUG,
        filename=log_file
    )
    logging.info("Successfully setup main app and settings")

    # Import each API and set it up
    app.register_blueprint(chru.api.web.v1.app, url_prefix="/api/v1")

    logging.info("Successfully setup and registered all APIs")

    return True

@app.route("/index")
@app.route("/")
def index():
    """ Simply renders the index page. """
    return render_template("index.html", recaptcha_key=web.s.captcha_public_key)

@app.route("/<slug>/")
@app.route("/<slug>")
def slug_redirect(slug):
    """ Redirects to the appropriate slug, if it exists.

        :param slug: the shortened url slug
        :type slug: str
        :return: a redirection (to index or the shortened url)
    """
    if not web.slug.is_valid(slug):
        logging.debug("%s isn't valid", slug)
        return redirect(url_for("index"))
    if not web.slug.exists(slug):
        logging.debug("%s doesn't exist", slug)
        return redirect(url_for("index"))
    url = web.slug.url_from_slug(slug)
    web.slug.add_hit(slug, real_ip(), request.user_agent)
    return redirect(url)

@app.route("/<slug>/delete/")
@app.route("/<slug>/delete")
def slug_delete_nokey(slug):
    """ Redirects back to the index if no deletion key is given. """
    return redirect(url_for("index"))

@app.route("/<slug>/delete/<delete>")
def slug_delete(slug, delete):
    """ Will attempt to delete a slug, if given a correct deletion key.

        If an invalid key is given, simply redirects to the index.

        :param slug: the shortened url slug to delete
        :type slug: str
        :param delete: the deletion key to use
        :type delete: str
        :return: a redirection to the index
    """
    if not web.slug.is_valid(slug) or not web.slug.exists(slug):
        return redirect(url_for("index"))

    row = web.slug.to_row(slug)
    if delete != row["delete"]:
        logging.debug("%s gave invalid delete code ('%s')", real_ip(), delete)
        return redirect(url_for("index"))

    deleted = web.slug.delete(slug)

    if not delete:
        logging.error("Unable to delete slug %s", slug)
        flash("Something went wrong deleting {0}".format(
            url_for("slug_redirect", slug=row["short"], _external=True)
        ), "error")
    else:
        flash("Successfully deleted {0}".format(
            url_for("slug_redirect", slug=row["short"], _external=True)
        ), "success")
    return redirect(url_for("index"))

@app.route("/<slug>/stats")
def slug_stats(slug):
    """ Pulls up all related slug information, and provides it to
        the statistics template. This gives information about the
        amount of clicks, operating systems, browsers, when it
        got the most clicks, and more.

        If there are no clicks, this really won't return much
        of value, as there is nothing!

        :param slug: the shortened url slug to show stats for
        :type slug: str
        :return: the rendered stats page or a JSON dict for API calls
        :enum: 1, 6, 7
    """
    if not web.slug.is_valid(slug) or not web.slug.exists(slug):
        return redirect(url_for("index"))

    stats = web.slug.pull_stats(slug, url_for)
    
    return render_template("stats.html", stats=stats)

@app.route("/submit", methods=["GET"])
def submit_get():
    """ Redirects GET requests to /submit to the index. """
    return redirect(url_for("index"))

@app.route("/submit", methods=["POST"])
def submit():
    """ Handles URL submissions, validating URLs,
        creating the needed slug row, and providing
        information to the client via JSON.

        This will handle invalid captchas, overly
        long URLs, invalid URLs (if validation is on),
        and custom slug related problems.

        :return: a JSON dictionary with submission info

        :enum: 7
    """
    logging.debug("HELLO, NOT USING_API HERE")
    response = captcha.submit(
        request.form["recaptcha_challenge_field"],
        request.form["recaptcha_response_field"],
        web.s.captcha_private_key,
        real_ip()
    )
    # reCAPTCHA failed
    if not response.is_valid:
        return jsonify({
            "error": "true",
            "enum": 2,
            "message": "The captcha typed was incorrect."
        })

    # long_url, custom_url, give_delete
    url_long = request.form["long_url"]
    url_short = request.form["custom_url"]
    url_delete = request.form["give_delete"] if "give_delete" in request.form else False

    # URL too long
    if len(url_long) > web.s.soft_url_cap:
        return jsonify({
            "error": "true",
            "enum": 3,
            "message": "The URL you provided was actually too long to shrink!"
        })

    # Invalid URL
    if web.s.validate_urls:
        valid = utility.funcs.validate_url(
            url=url_long,
            regex=web.s.validate_regex,
            use_requests=web.s.validate_requests,
            fallback=web.s.validate_fallback,
            timeout=5
        )
        if not valid:
            return jsonify({
                "error": "true",
                "enum": 4,
                "message": "The URL provided was unable to be validated."
            })

    # Custom URL validation (slugs)
    custom_slug = False
    if len(url_short) > 0:
        # If this isn't False at the end,
        # we display the error.
        invalid_slug = False
        slug_ = url_short

        if not re.match("^[\w-]+$", slug_):
            invalid_slug = "That custom URL is invalid. Only alphanumeric characters, dashes and underscores, please!"
        elif len(slug_) > web.s._slug_max:
            invalid_slug = "That custom URL is too long! Less than 32 characters, please!"
        elif slug_.lower() in web.s.reserved:
            invalid_slug = "Sorry, that custom URL is reserved!"
        elif web.slug.exists(web.s._schema["char"] + slug_):
            invalid_slug = "Sorry, that custom URL already exists!"

        if invalid_slug:
            return jsonify({
                "error": "true",
                "enum": 5,
                "message": invalid_slug
            })

        custom_slug = web.s._schema["char"] + slug_

    # Yippee, it's valid!

    slug_, id_, delete_, url_, old_ = web.slug.url_to_slug(
        url_long,
        ip=real_ip(),
        slug=custom_slug
    )

    # xxxx -> http://wht.nt/slug/delete/xxxx
    delete_ = url_for("slug_delete", slug=slug_, delete=delete_, _external=True)
    logging.debug("delete_ + url_for(): %s", delete_)

    delete_ = delete_ if url_delete and not old_ else ""

    data = {
        "error": "false",
        "enum": 0,
        "message": "Successfully shrunk your URL!" if not old_ else "URL already shrunk, here it is!",
        "url": url_for("slug_redirect", slug=slug_, _external=True),
        "delete": delete_,
        "given": url_
    }

    logging.debug("jsonify(data): %s", data)
    return jsonify(data)


""" **chr** (coded under the name ``chru``) is a Python based URL shortening
    service which uses Flask as a front end, and pysqlw as the SQL
    backend, to interface with sqlite3.

    It can shrink billions of unique URLs with less than 6 characters,
    run in the background with no human interaction,
    and it can fly like a bird -- or is that Super Man?
"""
from .version import __version__, __supported__, codes


# -*- coding: utf-8 -*-
class struct:
    """ This class is just a simple way to convert a
        dictionary to an object, for easier interaction.

        Instead of ``o['this']``, you can do ``o.this``;
        Isn't it great?!
    """
    def __init__(self, **entries):
        self.__dict__.update(entries)

    def __repr__(self):
        return str(self.__dict__)


# -*- coding: utf-8 -*-

"""
    This module just stores various functions which don't
    really fit anywhere else in the app.
"""

import pysqlw

import chru.web as web

import requests
import re
import logging
from urllib import unquote as urllib_unquote

from flask import request, url_for as flask_url_for

def url_for(endpoint, **kwargs):
    """ Sneaky hack allows us to easily unescape
        the HTML in _external url_for calls, since
        they're only used to display slug_redirect().
    """
    if "_external" in kwargs:
        return urllib_unquote(flask_url_for(endpoint, **kwargs))
    else:
        return flask_url_for(endpoint, **kwargs)

def real_ip(ip_header="X-Real-IP"):
    """ This checks whether or not the ip_header is set,
        which it is for chr.so, because it's run behind
        nginx and cloudflare.

        You'll have to figure out how to set that header
        if and only if you're running behind cloudflare
        for your particular httpd, I can't help, sorry.

        This is done because for some reason Flask makes it
        really fucking hard to just change where it pulls the
        remote address from, so we have to resort to workarounds.
    """
    if request.headers.getlist(ip_header):
        ip = request.headers.getlist(ip_header)[0]
    else:
        ip = request.remote_addr
    return ip

def sql():
    """ Opens an sqlite connection, returning it for usage.
    """
    return pysqlw.pysqlw(db_type="sqlite", db_path=web.s.sql_path)

def date_strip_day(date_):
    """ Strips the day out of a date.. I think.
    """
    if not "/" in date_:
        return date_
    return date_.split("/")[1]

def verify_sql(force=False):
    """ Ensure the database file is made, and if not,
        create it and populate it with the schema.

        :param force: should we force table creation?
        :type force: bool
    """
    import os
    if os.path.isfile(web.s.sql_path) and not force:
        return True
    with sql() as s:
        s.wrapper.cursor.executescript(web.schema)
        s.wrapper.dbc.commit()

    if not force:
        add_first = web.slug.url_to_slug("https://github.com/plausibility/chr", slug=web.s._schema["char"] + "source")
        return add_first
    else:
        return True

def constant(f):
    """ Simple decorator for const class variables.

        For example:

        .. code-block:: python

            >>> class A:
            ...    @constant
            ...    foo = 9001
            ...
            >>> a = A()
            >>> a.foo
            9001
            >>> a.foo = 7
            (...)
            Some SyntaxError
            (...)

        :raises: SyntaxError
    """
    def fset(self, value):
        raise SyntaxError("this is a constant variable")
    def fget(self):
        return f()
    return property(fget, fset)

def validate_url(url, timeout=1, fallback=True, regex=None, use_requests=True):
    """ Checks whether or not a URL is valid.
        The URL can go through two stages, depending
        on the settings passed in.

        1. The URL can be validated by sending a HEAD request
           with the requests module, and verifying that the reply
           code (if any), is in a list of valid replies.
        2. The URL gets run against a regex, to verify that,
           structurally, the URL is atleast valid.

        If fallback is on, the URL will go through both steps (if
        the first fails), to verify the validity.

        Failure of both tests results in a validity failure, obviously.

        :param url: the URL we're testing the validity of.
        :type url: str
        :param timeout: the timeout to pass to the HEAD request. this can stop exploits
        :type timeout: int
        :param fallback: should we fallback to the regex if the request fails?
        :type fallback: bool
        :param regex: the regex we'll compare against
        :type regex: str
        :param use_requests: should we use requests first?
        :type use_requests: bool
        :return: whether or not the URL is valid
        :rtype: bool
    """
    if use_requests:
        try:
            logging.debug("Sending HEAD req to: %s", url)
            head = requests.head(url, timeout=timeout)
            logging.debug("Reply: %s", head.status_code)
            # These are some codes we're 200 OK with. (geddit?)
            valid_codes = [
                200, # OK
                204, # No Content
                301, # Moved Permanently
                302, # Found
                304, # Not Modified
                307, # Temporary Redirect (HTTP/1.1)
                308, # Permanent Redirect (HTTP/1.1)
                418, # I'm a teapot
            ]
            if head.status_code not in valid_codes:
                raise requests.exceptions.HTTPError
            return True
        except requests.exceptions.RequestException as e:
            logging.debug(e)
            if fallback:
                return validate_url(url, timeout=timeout, fallback=False, regex=regex, use_requests=False)
            else:
                return False
    else:
        if regex is None:
            regex = web.s.validate_regex
        logging.debug("Matching '%s' against '%s'", url, regex)
        return re.match(regex, url, flags=re.I)


__version__ = "2.1.0"
__supported__ = ["2.1.0rc1", "2.1.0"]

import requests.structures

_codes = {
    -1:("not_post",),
    0: ("ok", "good", "success",),
    1: ("api_disabled", "disabled",),
    2: ("captcha_incorrect",),
    3: ("url_too_long",),
    4: ("url_invalid",),
    5: ("custom_url_invalid",),
    6: ("partial_form_data",),
    7: ("no_such_key",),
    8: ("no_such_url",),
    9: ("deletion_invalid",),
    10:("deletion_failed",)
}

codes = requests.structures.LookupDict(name="reply_codes")
for (code, titles) in list(_codes.items()):
    for title in titles:
        setattr(codes, title, code)
        if not title.startswith('\\'):
            setattr(codes, title.upper(), code)


#!/usr/bin/env python

"""
    Note: Taken from the public domain (namely: Stack Overflow), so this
    isn't my documentation.

    Converts any integer into a base [BASE] number. I have chosen 62
    as it is meant to represent the integers using all the alphanumeric
    characters, [no special characters] = {0..9}, {A..Z}, {a..z}
 
    I plan on using this to shorten the representation of possibly long ids,
    a la url shortenters
 
    saturate()  takes the base 62 key, as a string, and turns it back into an integer
    dehydrate() takes an integer and turns it into the base 62 string
"""
import math
import sys

BASE = 62

UPPERCASE_OFFSET = 55
LOWERCASE_OFFSET = 61
DIGIT_OFFSET = 48

NEGATIVE_BUFFER = '-'

def true_ord(char):
    """
    Turns a digit [char] in character representation
    from the number system with base [BASE] into an integer.
    """
    
    if char.isdigit():
        return ord(char) - DIGIT_OFFSET
    elif 'A' <= char <= 'Z':
        return ord(char) - UPPERCASE_OFFSET
    elif 'a' <= char <= 'z':
        return ord(char) - LOWERCASE_OFFSET
    else:
        raise ValueError("%s is not a valid character" % char)

def true_chr(integer):
    """
    Turns an integer [integer] into digit in base [BASE]
    as a character representation.
    """
    if integer < 10:
        return chr(integer + DIGIT_OFFSET)
    elif 10 <= integer <= 35:
        return chr(integer + UPPERCASE_OFFSET)
    elif 36 <= integer < 62:
        return chr(integer + LOWERCASE_OFFSET)
    else:
        raise ValueError("%d is not a valid integer in the range of base %d" % (integer, BASE))


def saturate(key):
    """
    Turn the base [BASE] number [key] into an integer
    Handles negatives by returning the negative.
    """
    if key[0] == NEGATIVE_BUFFER:
        return -saturate(key[1:])

    int_sum = 0
    reversed_key = key[::-1]
    for idx, char in enumerate(reversed_key):
        int_sum += true_ord(char) * int(math.pow(BASE, idx))
    return int_sum


def dehydrate(integer):
    """
    Turn an integer [integer] into a base [BASE] number
    in string representation

    Handles negatives by prefixing with a special
    character [NEGATIVE_BUFFER] and using the positive.
    """
    # handle negatives by prepending a plus and
    # using the positive number.
    if integer < 0:
        return NEGATIVE_BUFFER + dehydrate(-integer)

    # we won't step into the while if integer is 0
    # so we just solve for that case here
    if integer == 0:
        return '0'
    
    string = ""
    while integer > 0:
        remainder = integer % BASE
        string = true_chr(remainder) + string
        integer /= BASE
    return string


""" The chr api is a neat interface to a neat program.

    Lean, mean, and straight to the point, it aims to get work **done**.

    In and out in 60 seconds is the idea;
    we keep it simple and easy to work with.
"""


# -*- coding: utf-8 -*-

"""
    chra is the default entry-point for chr.
    Specifically, chru.chra:main is the entry point we use.
"""

from version import __version__, __supported__

import argparse
import os
import sys

import mattdaemon

class daemon(mattdaemon.daemon):
    def run(self, *args, **kwargs):
        """ This method will simply import the web control,
            set up the Flask app, and then run the app.
        """
        import web
        import web.routes
        setup = web.routes.set_app(
            debug=kwargs.get("debug"),
            settings_file=kwargs.get("settings")[0],
            log_file=kwargs.get("log"),
            __supported__=__supported__,
            update_schema=kwargs.get("update_schema")
        )
        if setup:
            web.routes.app.run(host=web.s.flask_host, port=web.s.flask_port)

def main():
    """ This method is in charge of everything

        It uses argparse to control the lovely arguments it handles,
        and sets up -- and controls -- the chr daemon.
    """
    if "--make-config" in sys.argv:
        import json
        import web
        print json.dumps(web.skeleton, indent=4)
        return

    parser = argparse.ArgumentParser(description="Take control of the chr url shortener.")
    parser.add_argument("action",
                        help="Action to take on the server",
                        choices=["start", "stop", "restart", "status", "-"],
                        metavar="action")

    parser.add_argument("-v", "--version",
                        help="Print the chr version and exit",
                        action="version",
                        version=__version__)

    parser.add_argument("-n", "--no-daemon",
                        help=argparse.SUPPRESS,
                        action="store_true")

    parser.add_argument("-l", "--log",
                        help="Set the file that output is logged to (default: %(default)s)",
                        type=str,
                        nargs=1,
                        metavar="file",
                        default="/tmp/chr-daemon.log")

    parser.add_argument("-p", "--pid",
                        help="Set the file the PID is saved to (default: %(default)s)",
                        type=str,
                        nargs=1,
                        metavar="file",
                        default="/tmp/chr-daemon.pid")

    parser.add_argument("-s", "--settings",
                        help="Set the location of the settings.json file",
                        type=file,
                        nargs=1,
                        metavar="file")

    # Not actually used via argparse.
    parser.add_argument("--make-config",
                        help="Print the settings.json file skeleton to stdout",
                        action="store_true")

    parser.add_argument("--update-schema",
                        help="Forcefully update the database schema, creating missing tables",
                        action="store_true")

    parser.add_argument("--api-manage",
                        help="Jump into the CLI API key manager",
                        action="store_true")

    parser.add_argument("--api-dump",
                        help="Dump the API keys as a JSON dict to stdout",
                        action="store_true")

    args = parser.parse_args()

    if args.api_manage:
        if not args.settings:
            print "You must pass in the settings.json file location (see --help for info)"
            return
        import api.manager
        cli = api.manager.cli(args.settings[0])
        cli.main()
        return

    if args.api_dump:
        if not args.settings:
            print "You must pass in the settings.json file location (see --help for info)"
            return
        import api.manager
        dump = api.manager.dump(args.settings[0])
        dump.dump()
        return

    daem_args = {
        "stdin": "/dev/null",
        "stdout": args.log,
        "stderr": args.log
    }
    if args.no_daemon:
        daem_args = {
            "stdin": "/dev/stdin",
            "stdout": "/dev/stdout",
            "stderr": "/dev/stderr"
        }
    daem = daemon(args.pid, daemonize=not args.no_daemon, **daem_args)

    if args.action == "start":
        if not args.settings:
            print "You must pass in the settings.json file location (see --help for info)"
            return

        # -n flag given? -- enable debugging
        debug = args.no_daemon
        daem.start(debug=debug, log=args.log, settings=args.settings, update_schema=args.update_schema)
    elif args.action == "stop":
        daem.stop()
    elif args.action == "restart":
        daem.restart()
    elif args.action == "status":
        if daem.status():
            print "chr currently running! :D"
        else:
            print "chr not running! D:"
        return
    elif args.action == "-":
        pass


if __name__ == "__main__":
    main()


import collections
import chru

s = None

# Under no circumstances should you touch this variable.
schema = """
-- THIS IS FOR sqlite3, NOT MySQL, SO DON'T TRY TO EVEN USE IT, BRO.

CREATE TABLE IF NOT EXISTS `users` (
    `id` INTEGER PRIMARY KEY,
    `login` varchar(32) NOT NULL,    -- Login name (lower case)
    `display` varchar(32) NOT NULL,  -- Displayed login name
    `password` varchar(64) NOT NULL, -- sha256() hash of (salt + login.lower() + password + salt)
    `email` varchar(64) NOT NULL,
    `admin` tinyint(1) NOT NULL DEFAULT '0'
);


-- To calculate hits: 'SELECT `id` FROM `clicks` WHERE `url` = <redirect_id>', then count.
CREATE TABLE IF NOT EXISTS `redirects` (
    `id` INTEGER PRIMARY KEY,
    `long` TEXT NOT NULL,
    `short` TEXT NOT NULL, -- Added afterwards. This is the base62 encoding for the row id.
    `delete` varchar(64) NOT NULL, -- String to delete. /slug/delete/<`delete`>
    `user` INTEGER NOT NULL DEFAULT '0', -- Which user made this? 0 means non-logged in.
    `ip` varchar(15) NOT NULL -- raw ip of the submitter.
);

CREATE TABLE IF NOT EXISTS `clicks` (
    `id` INTEGER PRIMARY KEY,
    `url` INTEGER NOT NULL,    -- id from redirect row
    `ip` varchar(15) NOT NULL, -- raw ip in form of 'W{1,3}.X{1,3}.Y{1,3}.Z{1,3}'
    `time` INTEGER NOT NULL,   -- int(time.time()) of hit
    `agent` TEXT NOT NULL,     -- Their user agent. Trimmed to a max of 256
    `agent_platform` TEXT NOT NULL, -- These are all based off of the
    `agent_browser` TEXT NOT NULL,  -- flask request.user_agent variable.
    `agent_version` TEXT NOT NULL,  -- It has the platform (Win, Mac), browser (Firefox, Safari)
    `agent_language` TEXT NOT NULL  -- aswell as version and language.
);

CREATE TABLE IF NOT EXISTS `api` (
    `id` INTEGER PRIMARY KEY,
    `key` TEXT NOT NULL,  -- The API key.
    `name` TEXT NOT NULL, -- The name of the key owner.
    `email` TEXT NOT NULL -- The email of the key owner.
);

"""

# Hack to make --make-config easier, OrderedDict means there's less problems.
# Reminder: change here, change in the docs!
skeleton = collections.OrderedDict()

skeleton["sql_path"] = "/path/to/chr.db"
skeleton["soft_url_cap"] = 512

skeleton["flask_host"] = "127.0.0.1"
skeleton["flask_port"] = 8080
skeleton["flask_base"] = "/"
skeleton["flask_secret_key"] = "UNIQUE_KEY"

skeleton["reserved"] = ["porn", "faggot", "sex", "nigger", "fuck",
                        "cunt", "dick", "gay", "vagina", "penis",
                        "tits", "boobs", "lesbian", "squirt",
                        "titties", "squirting", "shit", "anal"]

skeleton["captcha_public_key"] = "YOUR_PUBLIC_API_KEY"
skeleton["captcha_private_key"] = "YOUR_PRIVATE_API_KEY"

skeleton["salt_password"] = "UNIQUE_KEY"

skeleton["validate_urls"] = True
skeleton["validate_requests"] = True
skeleton["validate_regex"] = r"^(https?:\/\/)?([\da-z\.-]+)\.([a-z\.]{2,6})([\/\w \.!&()=:;?,\-%#]*)\/?$"
skeleton["validate_fallback"] = True

skeleton["api_enabled"] = False

# Stores DB table names (easly edet ur dotobuzzes)
skeleton["_schema"] = {
    "redirects": "redirects",
    "users": "users",
    "clicks": "clicks",
    "api": "api",
    "char": "+"
}

skeleton["_slug_max"] = 32

skeleton["_version"] = chru.__version__


# -*- coding: utf-8 -*-

import functools
import logging
import re

from flask import Blueprint, Flask, render_template, session, abort, request
from flask import g, redirect, url_for, flash, make_response, jsonify

import chru
import chru.web
import chru.utility.funcs as funcs

app = Blueprint('api_v1', __name__, template_folder="templates")

# Some constants that we'll use lots.
def _api_disabled(): # 1
    return {
        "error": "true",
        "enum": chru.codes.api_disabled,
        "message": "The chr API is disabled"
    }

def _url_too_long(): # 3
    return {
        "error": "true",
        "enum": chru.codes.url_too_long,
        "message": "The supplied URL was actually TOO long to shorten!"
    }

def _url_invalid(): # 4
    return {
        "error": "true",
        "enum": chru.codes.url_invalid,
        "message": "The supplied URL was unable to be validated."
    }

def _custom_url_invalid(msg): # 5
    return {
        "error": "true",
        "enum": chru.codes.custom_url_invalid,
        "message": msg
    }

def _partial_form_data(missing): # 6
    return {
        "error": "true",
        "enum": chru.codes.partial_form_data,
        "message": "You left fields: '{0}' out of your form".format("', '".join(missing))
    }

def _no_such_key(): # 7
    return {
        "error": "true",
        "enum": chru.codes.no_such_key,
        "message": "That API key does not exist"
    }

def _no_such_url(): # 8
    return {
        "error": "true",
        "enum": chru.codes.no_such_url,
        "message": "That shortened URL does not exist"
    }

# Decorators to make life easy as pie for us.

def requires_api():
    """ Decorates a function so that it will return
        an error if the chr API has been disabled.
    """
    def decorator(wrap_function):
        def wrapped_function(*args, **kwargs):
            if not chru.web.s.api_enabled:
                return jsonify(_api_disabled())
            return wrap_function(*args, **kwargs)
        return functools.update_wrapper(wrapped_function, wrap_function)
    return decorator

def requires_key():
    """ Decorates a function so that if the given
        api key is invalid, or there is no key
        given, it will return an error.
    """
    def decorator(wrap_function):
        def wrapped_function(*args, **kwargs):
            if "api_key" not in request.form:
                return jsonify(_no_such_key())
            api_key = request.form["api_key"]
            rows = funcs.sql().where("key", api_key).get(chru.web.s._schema["api"])
            if len(rows) != 1:
                return jsonify(_no_such_key())
            return wrap_function(*args, **kwargs)
        return functools.update_wrapper(wrapped_function, wrap_function)
    return decorator    

def requires_data(required):
    """ Decorates a function so that required
        form data is not present, it will
        return an error.
    """
    required = required + ["api_key"]
    def decorator(wrap_function):
        def wrapped_function(*args, **kwargs):
            missing = []
            for req in required:
                if req in request.form:
                    continue
                missing.append(req)
            if missing:
                return jsonify(_partial_form_data(missing))
            return wrap_function(*args, **kwargs)
        return functools.update_wrapper(wrapped_function, wrap_function)
    return decorator

# ----- DECORATORS

# !NB!
# You should put @app.route() calls BEFORE
# you call our @requires_*() decorators.
# The other way around means that Flask will
# directly bypass our wrapping, which is bad.

@app.route("/testing", methods=["POST"])
@requires_api()
@requires_key()
@requires_data(["hello", "test"])
def testing():
    return jsonify(request.form)

@app.route("/delete")
@app.route("/stats")
@app.route("/submit")
@requires_api()
def api_req_get():
    return jsonify({
        "error": "true",
        "enum": chru.codes.not_post,
        "message": "Please send all API requests as POST"
    })

@app.route("/submit", methods=["POST"])
@requires_api()
@requires_key()
@requires_data(["long", "short", "delete"])
def submit():
    url_long = request.form["long"]
    url_short = request.form["short"]
    url_delete = request.form["delete"]

    if chru.web.s.validate_urls:
        valid = funcs.validate_url(
            url=url_long,
            regex=chru.web.s.validate_regex,
            use_requests=chru.web.s.validate_requests,
            fallback=chru.web.s.validate_fallback,
            timeout=5
        )
        if not valid:
            return jsonify(_url_invalid())

    custom_slug = False
    if len(url_short) > 0:
        invalid_slug = False

        if not re.match("^[\w-]+$", url_short):
            invalid_slug = "That custom URL is invalid. Only alphanumeric characters, dashes and underscores, please!"
        elif len(url_short) > chru.web.s._slug_max:
            invalid_slug = "That custom URL is too long! Less than 32 characters, please!"
        elif url_short.lower() in chru.web.s.reserved:
            invalid_slug = "Sorry, that custom URL is reserved!"
        elif chru.web.slug.exists(chru.web.s._schema["char"] + url_short):
            invalid_slug = "Sorry, that custom URL already exists!"

        if invalid_slug:
            return jsonify(_custom_url_invalid(invalid_slug))

        custom_slug = chru.web.s._schema["char"] + url_short

    # Yippee, it's valid!

    slug_, id_, delete_, url_, old_ = chru.web.slug.url_to_slug(
        url_long,
        ip=funcs.real_ip(),
        slug=custom_slug
    )

    delete_ = delete_ if url_delete and not old_ else ""

    data = {
        "error": "false",
        "enum": chru.codes.ok,
        "message": "Successfully shrunk your URL!" if not old_ else "URL already shrunk, here it is!",
        "url": funcs.url_for("slug_redirect", slug=slug_, _external=True),
        "short": slug_,
        "delete": delete_,
        "given": url_
    }
    return jsonify(data)

@app.route("/stats", methods=["POST"])
@requires_api()
@requires_key()
@requires_data(["short"])
def stats():
    url_short = request.form["short"]
    if not chru.web.slug.is_valid(url_short) or not chru.web.slug.exists(url_short):
        return jsonify(_no_such_url())

    stats = chru.web.slug.pull_stats(url_short, funcs.url_for)
    stats = dict({
        "error": "false",
        "enum": chru.codes.ok,
        "message": "Here are your stats, piping hot from the oven!"
    }.items() + stats.items())

    return jsonify(stats)

@app.route("/delete", methods=["POST"])
@requires_api()
@requires_key()
@requires_data(["short", "delete"])
def delete():
    url_short = request.form["short"]
    url_delete = request.form["delete"]
    if not chru.web.slug.is_valid(url_short) or not chru.web.slug.exists(url_short):
        return jsonify(_no_such_url())

    row = chru.web.slug.to_row(url_short)
    if url_delete != row["delete"]:
        logging.debug("%s gave invalid delete code: %s", funcs.real_ip(), url_delete)
        return jsonify({
            "error": "true",
            "enum": chru.codes.deletion_invalid,
            "message": "The given delete code was invalid"
        })

    deleted = chru.web.slug.delete(url_short)

    url = funcs.url_for("slug_redirect", slug=row["short"], _external=True)

    if deleted:
        return jsonify({
            "error": "false",
            "enum": chru.codes.ok,
            "message": "Successfully deleted {0}".format(url)
        })
    else:
        return jsonify({
            "error": "true",
            "enum": chru.codes.deletion_failed,
            "message": "Unable to delete {0}".format(url)
        })

@app.route("/expand", methods=["POST"])
@requires_api()
@requires_key()
@requires_data(["short"])
def expand():
    url_short = request.form["short"]
    if not chru.web.slug.is_valid(url_short) or not chru.web.slug.exists(url_short):
        return jsonify(_no_such_url())

    row = chru.web.slug.to_row(url_short)

    return jsonify({
        "error": "false",
        "enum": chru.codes.ok,
        "message": "Here's the expanded URL",
        "long": row["long"],
        "short": row["short"]
    })


# -*- coding: utf-8 -*-

from chru.version import __supported__

import json
import pysqlw

class cli:
    """ A really nasty and horrible CLI implementation
        of an API key manager. I should probably improve
        it some day, but it'll work for now.
    """

    def __init__(self, settings):
        self.settings = json.load(settings)
        if not "_version" in self.settings \
                or self.settings["_version"] not in __supported__:
            raise ValueError("The given settings file is not supported.")
        settings.close()

    def sqlw(self):
        return pysqlw.pysqlw(db_type="sqlite", db_path=self.settings["sql_path"])

    def menu(self):
        items = [(1, "List API keys"),
                 (2, "Add a new key"),
                 (3, "Delete a key")]
        print "What do you want to do? (Ctrl-C to exit)"

        for k, v in items:
            print str(k) + '.', v

        try:
            option = int(raw_input("# "))
        except ValueError as e:
            print "That's not a number... Try again."
            return self.menu()

        for k, v in items:
            if option == k:
                return option

        print "That's no good, try again.."
        return self.menu()

    def api_list(self):
        rows = self.sqlw().get(self.settings["_schema"]["api"])
        print ""
        if len(rows) > 0:
            print "List of API keys:"
            for row in rows:
                print "#{0}: {1} ({2} {3})".format(row["id"], row["key"], row["name"], row["email"])
        else:
            print "No API keys were found! :("

    def api_add(self):
        print "\nPlease provide info about API key owner:"
        correct_info = False
        owner = False
        email = False
        key = self.make_random()
        while not correct_info:
            owner = str(raw_input(" name: "))
            email = str(raw_input("email: "))
            print ":: {0} ({1})\n".format(owner, email)
            correct = str(raw_input("Is this correct? [Y/n] ")).lower()
            correct_info = correct in ("", "y", "ye", "yes", "yup", "correct-a-mundo")
        data = {
            "key": key,
            "name": owner,
            "email": email
        }
        if self.sqlw().insert(self.settings["_schema"]["api"], data):
            print "Successfully added API key!"
            print ":: ", key

    def api_delete(self):
        print "\nWhat number is the key? (As provided by the listing): ('q' to go back)"
        num = False
        while not num:
            try:
                n = raw_input("># ")
                if n.lower() in ("q", "quit"):
                    return
                num = int(n)
            except ValueError as e:
                print "That isn't a number, try again."
        if len(self.sqlw().where("id", num).get(self.settings["_schema"]["api"])) != 1:
            print "That API key doesn't seem to exist, sorry."
            return

        if self.sqlw().where("id", num).delete(self.settings["_schema"]["api"]):
            print "Successfully deleted API key #{0}!".format(num)
        else:
            print "Unable to delete API key. :("


    def make_random(self, length=32):
        import string, random
        parts = string.uppercase + string.lowercase + string.digits
        return "".join(random.choice(parts) for x in range(length))

    def main(self):
        funcs = {
            1: self.api_list,
            2: self.api_add,
            3: self.api_delete
        }
        try:
            while True:
                option = self.menu()
                funcs[option]()
                print ""
        except KeyboardInterrupt as e:
            print ""
            print "Goodbye!"

class dump:
    def __init__(self, settings):
        self.settings = json.load(settings)
        if not "_version" in self.settings \
                or self.settings["_version"] not in __supported__:
            raise ValueError("The given settings file is not supported.")
        settings.close()

    def sqlw(self):
        return pysqlw.pysqlw(db_type="sqlite", db_path=self.settings["sql_path"])

    def dump(self):
        rows = self.sqlw().get(self.settings["_schema"]["api"])
        data = { "keys": [] }
        for row in rows:
            data["keys"].append(row)
        print data


# -*- coding: utf-8 -*-

"""
    chru.web.slug provides an easy to use interface to the
    shortened URL slugs, allowing creation, removal, editing, and more.

    These functions depend on the SQL database, so without it they're useless.
"""

from datetime import datetime
import time
import re
import logging

import chru.utility as utility
import chru.utility.base62 as base62
import chru.web as web

def pull_stats(slug, url_for):
    """ Cluster together the statistics for a
        given slug. This pulls out information
        such as:

        - Number of clicks
        - Info about the clicks
          - Browser
          - Platform
          - Clicks per day
        - Unique/return clicks

        :param slug: the slug we're gathering statistics for
        :type slug: str
        :param url_for: the url_for() function we can use to expand urls
        :type url_for: def
        :return: statistics about a url
        :rtype: dict
    """
    row = to_row(slug)
    clicks = utility.funcs.sql().where("url", row["id"]).get(web.s._schema["clicks"])

    click_info = {
        "platforms": {"unknown": 0},
        "browsers": {"unknown": 0},
        "pd": {} # clicks per day
    }
    month_ago = int(time.time()) - (60 * 60 * 24 * 30) # Last 30 days
    month_ago_copy = month_ago

    for x in xrange(1, 31):
        # Pre-populate the per day clicks dictionary.
        # This makes life much much easier for everything.
        month_ago_copy += (60 * 60 * 24) # one day
        strf = str(datetime.fromtimestamp(month_ago_copy).strftime("%m/%d"))
        click_info["pd"][strf] = 0

    for click in clicks:
        if click["time"] > month_ago:
            # If the click is less than a month old
            strf = str(datetime.fromtimestamp(click["time"]).strftime("%m/%d"))
            click_info["pd"][strf] += 1

        if len(click["agent"]) == 0 \
                or len(click["agent_platform"]) == 0 \
                or len(click["agent_browser"]) == 0:
            click_info["platforms"]["unknown"] += 1
            click_info["browsers"]["unknown"] += 1
            continue

        platform_ = click["agent_platform"]
        browser_ = click["agent_browser"]

        if not platform_ in click_info["platforms"]:
            click_info["platforms"][platform_] = 1
        else:
            click_info["platforms"][platform_] += 1

        if not browser_ in click_info["browsers"]:
            click_info["browsers"][browser_] = 1
        else:
            click_info["browsers"][browser_] += 1

    # Formatting in directly without escapes, how horrible am I!
    # But these are known to be *safe*, as they're set by us.
    unique = utility.funcs.sql().query("SELECT COUNT(DISTINCT `{0}`) AS \"{1}\" FROM `{2}`".format("ip", "unq", web.s._schema["clicks"]))
    unique = unique[0]["unq"]

    rpt = len(clicks) - unique
    if len(clicks) == 0:
        rpt = 0

    f_unique = unique
    f_clicks = len(clicks) if len(clicks) != 0 else 1

    hits = {
        "all": len(clicks),
        "unique": unique,
        "return": rpt,
        # (unique / all)% of visits only come once
        "ratio": str(round(float(f_unique) / float(f_clicks), 2))[2:]
    }

    stats = {
        "hits": hits,
        "clicks": click_info,
        "long": row["long"],
        "long_clip": row["long"],
        "short": row["short"],
        "short_url": url_for("slug_redirect", slug=row["short"], _external=True)
    }

    if len(row["long"]) > 30:
        stats["long_clip"] = row["long"][:30] + "..."

    return stats

def url_to_slug(url, ip=False, slug=False):
    """ Shrinks a URL to the slug equivalent.

        No URL length limits are imposed, these should
        be done **before** calling.

        :param url: the long URL to shrink
        :type url: str
        :param ip: the IP address of the creator
        :type ip: str or False
        :param slug: the custom slug (if any) to use. If not given, one is created.
        :type slug: str or False
        :return: tuple with (shortened slug, row id, deletion key, long url, old or new url)
        :rtype: tuple
    """
    old_slug = utility.funcs.sql().where("long", url).get(web.s._schema["redirects"])
    if len(old_slug) > 0:
        # url already shortened, give that.
        old_slug = old_slug[0]
        return (old_slug["short"], old_slug["id"], old_slug["delete"], old_slug["long"], True,)
    data = {
        "long": url,
        "short": "",
        "delete": make_delete_slug(),
        "ip": ip if ip else ""
    }
    sq = utility.funcs.sql()
    row = sq.insert(web.s._schema["redirects"], data)
    if not row:
        logging.debug("Unable to insert: %s", data)
        return False
    id = sq.wrapper.cursor.lastrowid
    if not slug:
        logging.debug("Got %s from slug add: %s", id, url)
        slug = id_to_slug(id)
    update = {
        "short": slug
    }
    if utility.funcs.sql().where("id", id).update(web.s._schema["redirects"], update):
        slug = utility.funcs.sql().where("id", id).get(web.s._schema["redirects"])[0]
        return (slug["short"], slug["id"], slug["delete"], slug["long"], False,)
    else:
        return False

### conversions

def slug_to_id(slug):
    """ Turns a short slug into the row ID equivalent, using base62.

        For custom slugs, this is simply done by querying the database.

        :param slug: the shortened slug to convert
        :type slug: str
        :return: the slug's ID, if any.
        :rtype: int
        :raises: ValueError
    """
    if slug[:len(web.s._schema["char"])] == web.s._schema["char"]:
        rows = utility.funcs.sql().where("short", slug).get(web.s._schema["redirects"])
        if len(rows) != 1:
            return False
        return rows[0]["id"]

    if not is_valid(slug):
        return False
    if not isinstance(slug, basestring):
        raise ValueError("Wanted str, got " + str(type(slug)))
    return base62.saturate(slug)

def id_to_slug(id):
    """ Dehydrates an integer into the string equivalent.

        :param id: the id to saturate
        :type id: int
        :return: the saturated slug, if any
        :rtype: int
        :raises: ValueError
    """
    if not isinstance(id, int):
        raise ValueError("Wanted int, got " + str(type(id)))
    return base62.dehydrate(id)

def to_row(slug):
    """ Returns all database row information relating to a slug.

        Custom slug info is found by querying the database.

        :param slug: the slug to pull the data for
        :type slug: str
        :return: the slug's database row (False if not found)
        :rtype: dict or False
    """
    if not is_valid(slug):
        return False

    if slug[:len(web.s._schema["char"])] == web.s._schema["char"]:
        rows = utility.funcs.sql().where("short", slug).get(web.s._schema["redirects"])
    else:
        rows = utility.funcs.sql().where("id", slug_to_id(slug)).get(web.s._schema["redirects"])
    return rows[0] if len(rows) == 1 else False

### info

def hits(slug):
    """ Return the number of clicks a slug has had.
        This is done by pulling all the clicks from the database,
        then counting them.

        :param slug: the slug to count clicks for
        :type slug: str
        :return: the number of clicks the slug has
        :rtype: int
    """
    if not is_valid(slug):
        return 0
    id = slug_to_id(slug)
    rows = utility.funcs.sql().where("url", id).get(web.s._schema["clicks"])
    return len(rows)

def exists(slug):
    """ Check if a given slug actually exists.
        This will just query the database and verify that there is only one row.

        :param slug: the slug to check the existance of
        :type slug: str
        :return: if the slug exists
        :rtype: bool
    """
    if not is_valid(slug):
        return False
    if slug[:len(web.s._schema["char"])] == web.s._schema["char"]:
        rows = utility.funcs.sql().where("short", slug).get(web.s._schema["redirects"])
    else:
        rows = utility.funcs.sql().where("id", slug_to_id(slug)).get(web.s._schema["redirects"])
    return len(rows) == 1

def is_valid(slug):
    """ Attempts to saturate a slug, to check validity.
        Custom slugs are assumed to be valid, unless otherwise shown.

        :param slug: the slug to check the validity of
        :type slug: str
        :return: whether the slug is valid
        :rtype: bool
    """
    if slug[:len(web.s._schema["char"])] == web.s._schema["char"]:
        return True

    try:
        base62.saturate(slug)
        return True
    except (TypeError, ValueError) as e:
        logging.debug(e)
        return False

def url_from_slug(slug):
    """ Find the long URL a slug is hiding.
        This is simply found by querying the database for the data.

        :param slug: the slug to expand to a long url
        :type slug: str
        :return: the expanded URL (or False if invalid)
        :rtype: str or bool
    """
    if not is_valid(slug):
        return False
    if not exists(slug):
        return False
    if slug[:len(web.s._schema["char"])] == web.s._schema["char"]:
        row = utility.funcs.sql().where("short", slug).get(web.s._schema["redirects"])[0]
    else:
        row = utility.funcs.sql().where("id", slug_to_id(slug)).get(web.s._schema["redirects"])[0]
    return row["long"]

### modification

def add_hit(slug, ip, ua, time_=None):
    """ Add a click for the given slug to the database.
        Stores the user agent info, as well as IP address and timestamp.

        :param slug: the slug to add a click for
        :type slug: str
        :param ip: the IP address of the clicking user
        :type ip: str
        :param ua: the Flask useragent object for the clicking user
        :type ua: Flask useragent object
        :param time\_: an optional unix timestamp to attach to the click (defaults to now)
        :type time\_: int or None
        :return: whether or not the insertation was successful
        :rtype: bool
    """
    if not is_valid(slug):
        return False
    if time_ is None:
        time_ = int(time.time())
    id = slug_to_id(slug)
    data = {
        "url": id,
        "ip": ip,
        "time": time_,
        # agent is max of 256 characters, for brevity.
        "agent": ua.string[:256] if not ua.string == None else "",
        "agent_platform": ua.platform if not ua.platform == None else "",
        "agent_browser": ua.browser if not ua.browser == None else "",
        "agent_version": ua.version if not ua.version == None else "",
        "agent_language": ua.language if not ua.language == None else ""
    }
    return utility.funcs.sql().insert(web.s._schema["clicks"], data)

def delete(slug):
    """ Delete the given slug entirely from the database.
        This will also delete any clicks attached to this slug, to save (a little) space.

        :param slug: the slug to delete
        :type slug: str
        :return: whether or not the deletions were successful
        :rtype: bool
    """
    if not is_valid(slug) or not exists(slug):
        return False
    id = slug_to_id(slug)
    rem_slug = utility.funcs.sql().where("id", id).delete(web.s._schema["redirects"])
    rem_click = utility.funcs.sql().where("url", id).delete(web.s._schema["clicks"])
    if not rem_slug or not rem_click:
        logging.debug("Unable to fully delete: %s, rem_slug: %s, rem_click: %s", slug, rem_slug, rem_click)
    return rem_slug

### generators

def make_delete_slug(length=64):
    """ Generate a deletion key for a slug.
        This is used incase the user wants to delete their URL.

        :param length: the length of the key
        :type length: int
        :return: a deletion key
        :rtype: str
    """
    import string, random
    parts = string.uppercase + string.lowercase + string.digits
    return "".join(random.choice(parts) for x in range(length))


from setuptools import setup

if __name__ != "__main__":
    import sys
    sys.exit(1)

def long_desc():
    with open('README.rst', 'rb') as f:
        return f.read()

execfile('chru/version.py')

kw = {
    "name": "chr",
    "version": __version__,
    "description": "Python based URL shortening service",
    "long_description": long_desc(),
    "url": "https://github.com/plausibility/chr",
    "author": "plausibility",
    "author_email": "chris@gibsonsec.org",
    "license": "MIT",
    "packages": [
        'chru',
        'chru.api',
        'chru.api.web',
        'chru.web',
        'chru.utility',
    ],
    "include_package_data": True,
    "install_requires": [
        "requests",
        "flask",
        "flask-kvsession",
        "recaptcha-client",
        "mattdaemon>=1.1.0",
        "pysqlw>=1.3.0"
    ],
    "zip_safe": False,
    "keywords": "url short shortener slug",
    "classifiers": [
        "Development Status :: 4 - Beta",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 2"
    ],
    "entry_points": {
        "console_scripts": [
            'chru = chru.chra:main'
        ]
    }
}

if __name__ == "__main__":
    setup(**kw)
